from setuptools import setup, find_packages
setup(name='GhostScan',
      version='0.5.25',
      description='three-dimensional imaging and projection ',
      url='https://github.com/NU-ACCESS/GhostScan',
      author='Frank Yang',
      author_email='frankyang2024@u.northwestern.edu',
      license='MIT',
      packages=find_packages(),
      install_requires=['numpy>=1.11',
                        'matplotlib>=1.5',
                        'Pillow >= 7.0.0',
                        'matplotlib >= 3.2.1',
                        'numpy >= 1.18.2',
                        'torch >= 1.4.0',
                        'tifffile >= 2020.6.3',
                        'scipy >= 1.5.2',
                        'wavepy >= 1.0.5',
                        'checkerboard >= 0.2.4',
                        'tqdm >= 4.48.2',
                        'opencv-contrib-python >= 4.4.0',
                        'selenium>=3.141.0',
                        'pandas>=1.1.1',
                        'scikit-image>=0.17.2',
                        'ipython>=7.18.1',
                        'bokeh>=0.12.5'])
